<?php

namespace EuKit\Base\Parts;

use EuKit\Base\Interfaces\ViewModel;
use EuKit\Base\SiteLayout;
use Infomaker\Everyware\NewsML\Objects\ImageItem;
use Infomaker\Everyware\Support\NewRelicLog;
use Infomaker\Imengine\CropData\AutoCropData;
use Infomaker\Imengine\CropData\CropResizeData;
use Infomaker\Imengine\Helpers\ImengineHelper;
use Infomaker\Imengine\Imengine;
use RuntimeException;

class Image implements ViewModel
{
    private const DEFAULT_SIZES_CSS_LENGTH = '100%';

    /**
     * @var ImageItem
     */
    private $image;

    /**
     * @var string
     */
    private $ratio = '16:9';

    /**
     * @var bool If FALSE, and no predefined crop size available in $this->ratio,
     *           use the image's original ratio rather than cropping automatically.
     */
    private $hasAutomaticCropEnabled = true;

    /**
     * @var string The maximum width the image can take on different screen sizes.
     *             In percentage relative to its container width, or in a CSS unit.
     *             Examples: '66.66%', '10em', '200px'
     */
    private $sizesData = self::DEFAULT_SIZES_CSS_LENGTH;

    /**
     * @var array Arguments to aid the generation of sizes; desscription of grid and content sizes
     */
    private $args = [];

    public function __construct(ImageItem $image)
    {
        $this->image = $image;

        $flags = $this->image->get('flags', []);
        foreach ($flags as $flag) {
            if ($flag['name'] === 'disableAutomaticCrop') {
                $this->hasAutomaticCropEnabled = false;
                break;
            }
        }
    }

    /**
     * @param array $args
     */
    public function setArgs(array $args): void
    {
        $defaults = [
            'max_cols'      => null,
        ];

        $this->args = array_merge($defaults, $args);
    }

    /**
     * @param string $ratio Examples: '16:9', '3:2'
     */
    public function setRatio(string $ratio): void
    {
        $this->ratio = $ratio;
    }

    /**
     * @param bool $enabled
     */
    public function setAutomaticCropEnabled(bool $enabled): void
    {
        $this->hasAutomaticCropEnabled = $enabled;
    }

    /**
     * @param string $sizesData
     */
    public function setSizesData(string $sizesData): void
    {
        $this->sizesData = $sizesData;
    }

    public function getRatioUrl($width): string
    {
        $height = $this->getHeight($width);

        try {

            if ($this->image->hasCropForRatio($this->ratio)) {

                // Use a predefined crop.
                $crop = $this->image->getCropForRatio($this->ratio);
                $crop_handler = new CropResizeData($crop['width'], $crop['height']);
                $crop_handler
                    ->setCropX($crop['x'])
                    ->setCropY($crop['y'])
                    ->setWidth($width)
                    ->setHeight($height);

                return Imengine::cropResize($crop_handler)->fromUuid($this->image->uuid);
            }

            if (!$this->hasAutomaticCropEnabled) {

                // Don't crop at all.
                if ($this->image->hasOriginalSizes()) {
                    $maxSize = max($width, $height);
                }
                else {
                    /**
                     * If the image is portait-shaped, this will make it too small.
                     * But at least it will not be cropped.
                     */
                    $maxSize = $width;
                }

                return Imengine::fit((int)$maxSize)->fromUuid($this->image->uuid);
            }

            if ($this->image->hasOriginalSizes()) {

                // Automatic crop with center/top align.
                $crop_handler = new AutoCropData($this->image->width, $height, $width);
                $crop_handler->setRatio($this->ratio);
                $align_top = ['y' => ImengineHelper::formatNumberToImengineFloat(0)];

                return Imengine::autoCrop($crop_handler)->with($align_top)->fromUuid($this->image->uuid);
            }

            // Hard crop.
            return Imengine::ratio($width, $this->ratio)->fromUuid($this->image->uuid);

        } catch (RuntimeException $e) {
            NewRelicLog::error('Failed to fetch imengine url', $e);
        }

        return '';
    }

    private function getHeight($width): ?int
    {
        if ($this->image->hasCropForRatio($this->ratio)) {

            return ImengineHelper::calculateRatioHeight($width, $this->ratio);
        }

        if (!$this->hasAutomaticCropEnabled) {

            if ($this->image->hasOriginalSizes()) {
                $scaling = $width / $this->image->width;

                return $this->image->height * $scaling;
            }

            // Height is unknown.
            return null;
        }

        if ($this->image->hasOriginalSizes()) {
            return $this->image->height;
        }

        // All else failed, so revert to hard crop.
        return ImengineHelper::calculateRatioHeight($width, $this->ratio);
    }

    public function getFitURL($width): string {
        try{
            return Imengine::fit($width)->fromUuid($this->image->uuid);
        } catch (RuntimeException $e){
            NewRelicLog::error('Failed to fetch imengine url', $e);
        }

        return '';
    }

    private function createSrcset(array $sizes, bool $crop = true): string
    {
        $srcset = [];
        foreach ($sizes as $size) {
            if (is_int($size)) {
                if ($crop){
                    $srcset[] = $this->getRatioUrl($size) . " {$size}w";
                }else{
                    $srcset[] = $this->getFitURL($size) . " {$size}w";
                }

            }
        }

        return implode(',', $srcset);
    }

    private function createSizes(): string
    {
        // No point in making a media query for each screen size, if they are all going to have the same fixed length.
        $evaluateEachScreenSize = is_float($this->getImageFractionOfGrid($this->sizesData));

        if (!$evaluateEachScreenSize) {
            return $this->getMediaQueryAndSize($this->sizesData);
        }

        $sizes = [];
        $imageWidth = $this->sizesData;
        $contentFractionOfGrid = 1.0;
        $i = -1;
        $breakpoints = SiteLayout::getContentSizeBreakpoints();
        foreach ($breakpoints as $screenSizeName => $minScreenWidth) {
            $i++;

            // Always define the smallest screen size.
            // Else, skip if there is no specific content size defined for it.
            $contentSize = $this->args['content_sizes'][$screenSizeName] ?? null;
            if ($contentSize === null && $i > 0) {
                continue;
            }

            $contentFractionOfGrid = ($contentSize !== null && $this->args['max_cols'] !== null)
                ? $contentSize / $this->args['max_cols'] : 1.0;

            $sizes[$minScreenWidth] = $this->getMediaQueryAndSize($imageWidth, $minScreenWidth, $contentFractionOfGrid);
        }

        // If a maximum page width is known, limit image width to the size it would have in its container at that width.
        if (SiteLayout::getMaxPageWidth() !== null) {
            $sizes[] = $this->getMediaQueryAndSize($imageWidth, SiteLayout::getMaxPageWidth(), $contentFractionOfGrid, true);
        }

        // The grid breakpoints are sorted in ascending size order, but in a `sizes` attribute we want to start with the largest media query.
        $sizes = array_reverse($sizes);

        return implode(',', $sizes);
    }

    /**
     * Get a media query / size pair (or just a size) for given screen size and content size.
     *
     * @param string $imageWidth            CSS length. Examples: '66.67%', '100px', '15vw'
     * @param int    $minScreenWidth        In pixels
     * @param float  $contentFractionOfGrid A positive float, maximum 1.0
     * @param bool   $fixFractionToPixelSize
     *
     * @return string Something like '(min-width: 800px) 33.33vw' or '66.66vw'
     */
    private function getMediaQueryAndSize(string $imageWidth, int $minScreenWidth = 0, float $contentFractionOfGrid = 1.0, bool $fixFractionToPixelSize = false): string
    {
        $mediaQuery = ($minScreenWidth > 0) ? sprintf('(min-width: %dpx) ', $minScreenWidth) : '';

        $imageWidthFraction = $this->getImageFractionOfGrid($imageWidth, $contentFractionOfGrid);
        if (is_float($imageWidthFraction)) {
            $size = ($fixFractionToPixelSize)
                ? sprintf('%0.2dpx', ($imageWidthFraction * $minScreenWidth))
                : sprintf('%0.2fvw', ($imageWidthFraction * 100));
        } else {
            $size = $imageWidth;
        }
        if (SiteLayout::getContainerHorizontalPadding() !== null) {
            $size = 'calc(' . $size . ' - (' . SiteLayout::getContainerHorizontalPadding() . '))';
        }

        return trim($mediaQuery . ' ' . $size);
    }

    /**
     * If $imageWidth is a percentage, return it as a fraction of the grid width. Else NULL
     *
     * @param string $imageWidth         CSS length. Examples: '66.67%', '100px', '15vw'
     * @param float  $fractionOfGridUsed A positive float, maximum 1.0
     *
     * @return float|null  A positive float, maximum 1.0
     */
    private function getImageFractionOfGrid(string $imageWidth, float $fractionOfGridUsed = 1.0): ?float
    {
        if (substr($imageWidth, -1) !== '%') {
            return null;
        }
        return (float)substr($imageWidth, 0, -1) / 100 * $fractionOfGridUsed;
    }

    public function getViewData(): array
    {
        $widths = [300, 600, 900, 1200, 1500];
        $defaultWidth = $this->image->width ?? max($widths);
        $defaultHeight = $this->getHeight($defaultWidth) ?? '';
        return array_replace($this->image->toArray(), [
            'authors' => $this->image->get('authors'),
            'defaultCrop' => [
                'width' => $defaultWidth,
                'height' => $defaultHeight,
            ],
            'src' => $this->getRatioUrl($defaultWidth),
            'fitSrc' => $this->getFitURL($defaultWidth),
            'srcset' => $this->createSrcset($widths),
            'fitSrcset' => $this->createSrcset($widths, false),
            'sizes' => $this->createSizes(),
        ]);
    }
}
